#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <bio.h>
#include <ndb.h>

/*	connect ver.1.1a	*/

int vflag = 0;
int Kflag = 0;
int tflag = 0;

void
usage(void)
{
	fprint(2,"usage: connect [-6vKt] addr\n"
	"	-6: IPv6 connection\n"
	"	-v: verbose\n"
	"	-K: Kill the process that is waitng for response from remote server\n"
	"	-t: no coversation. just dial test\n"
	"example: connect tcp!HOST!80\n");
	exits("usage");
}

void
kill(int pid)
{
	char buf[128];
	int fd;
	snprint(buf, sizeof(buf),"/proc/%d/note",pid);
	fd = open(buf,OWRITE);
	write(fd,"kill",4);
	close(fd);
}

int
xfer(int fd0, int fd1)
{
	char buf[16384];
	int n,m;
	while((n = read(fd0,buf, sizeof(buf))) > 0){
		m = write(fd1, buf, n);
		if(m < 0)
			break;
	}
	write(fd1,"",0);
	return (n <= 0)?0:1;
}


void
connect(int fd0, int fd1, char *dest)
{	int data;
	char devdir[NETPATHLEN];
	int pid;
	int status;

	if(vflag)
		fprint(2, "dialing to %s\n", dest);
	data = dial(dest, 0, devdir, 0);
	if(data < 0)
		sysfatal("%r");
	if(vflag || tflag){
		NetConnInfo *cinfo;
		fprint(2, "connected to %s on %s\n", dest, devdir);
		cinfo = getnetconninfo(nil, data);
		fprint(2, "remote %s\n", cinfo->raddr);
		fprint(2, "local  %s\n", cinfo->laddr);
	}
	if(tflag)
		exits(nil);

	switch((pid=rfork(RFPROC|RFFDG|RFMEM))){ /* assign = */
	case -1:
		perror("connect");
		exits("fork");
	case 0: // local to remote
		sleep(100); // wait for my partner to be ready
		status = xfer(fd0,data);
		if(vflag)
			fprint(2, status?"callee closed\n":"caller closed\n");
		if(Kflag)
			kill(getppid());
		exits(0);
	default: // remote to local
		status = xfer(data,fd1);
		if(vflag)
			fprint(2, status?"caller closed\n":"callee closed\n");
		kill(pid);
		exits(0);
	}
}

/*	stolen from ping.c	*/
static int
isdottedquad(char *name)
{
	int dot = 0, digit = 0;

	for (; *name != '\0'; name++)
		if (*name == '.')
			dot++;
		else if (isdigit(*name))
			digit++;
		else
			return 0;
	return dot && digit;
}

/*	stolen from ping.c	*/
static int
isv6lit(char *name)
{
	int colon = 0, hex = 0;

	for (; *name != '\0'; name++)
		if (*name == ':')
			colon++;
		else if (isxdigit(*name))
			hex++;
		else
			return 0;
	return colon;
}

char *	/* new addr */
mknetaddr(char *addr, int ipver)
/*	addr must be network!netaddr!service */
{
	int n;
	char *p,*q;
	Ndbtuple *t, *nt;
	char *netroot = nil;
	char *taddr;
	char *args[3];
	taddr = strdup(addr); // to be scratched
	n = getfields(taddr,args,3,0,"!");
	if(n != 3)
		goto ERR;
	if(isdottedquad(args[1]) || isv6lit(args[1]))
		return addr;

	q = csgetvalue("/net", "dom",args[1],"ip",&t);
	if(q == nil)
		sysfatal("csgetvalue %s: %r",args[1]);
	/* q is the first ip addr that is found by cs. 
	 * we donn't need this one.
	 * all ip are searchable using t. */

	if(*args[0] != '/')
		netroot = smprint("/net/%s",args[0]);
	p = nil;
	for(nt = t; nt != nil; nt = nt->entry) {
		if(vflag)
			print("%s=%s\n",nt->attr,nt->val);
		/* note: csgetvalue() returns all attr even if "ip" is specified */
		if(strcmp(nt->attr, "ip") == 0) {
			p = netmkaddr(nt->val, args[0], args[2]);
			q = strchr(p,'!');
			if(ipver == 4 && strchr(q,'.'))
				break;
			if(ipver == 6 && strchr(q,':'))
				break;
			/* don't free(p). look dial(2) */
		}
		p = nil;
	}
	if(!p)
		goto ERR;
	free(taddr);
	if(netroot)
		free(netroot);
	ndbfree(t);
	return p;
ERR:
	werrstr("bad address");
	free(taddr);
	return nil;
}

void
main(int argc, char *argv[])
{
	int ipver = 4;
	char *p;
	ARGBEGIN{
	case '6': ipver = 6; break;
	case 'v': vflag = 1; break;
	case 'K': Kflag = 1; break;
	case 't': tflag = 1; break;
	default: usage();
	}ARGEND

	if(argc != 1)
			usage();
	p = mknetaddr(argv[0],ipver);
	if(!p)
		sysfatal("%r");
	connect(0,1,p);
	exits(nil);
}
